@@ -1,16 +1,5 @@
 package Starlink::AST::Tk;
 
-use strict;
-use vars qw/ $VERSION /;
-use constant R2D     => 57.29578;        # Radians to degrees factor
-use constant FLT_MAX => 3.40282347e+38;  # Maximum float on ix86 platform
-
-use Tk;
-use Starlink::AST;
-use Carp;
-
-'$Revision: 1.15 $ ' =~ /.*:\s(.*)\s\$/ && ($VERSION = $1);
-
 =head1 NAME
 
 Starlink::AST::Tk - AST wrapper to the Tk library
@@ -48,11 +37,99 @@ of the packages public interface.
 
 =head1 REVISION
 
-$Id: Tk.pm,v 1.15 2004/04/22 02:19:00 timj Exp $
+$Id: Tk.pm,v 1.17 2005/03/04 19:22:44 timj Exp $
+
+=cut
+
+use strict;
+use vars qw/ $VERSION /;
+use constant R2D     => 57.29578;        # Radians to degrees factor
+use constant FLT_MAX => 3.40282347e+38;  # Maximum float on ix86 platform
+
+use Tk;
+use Tk::Font;
+use Starlink::AST;
+use Carp;
+
+'$Revision: 1.17 $ ' =~ /.*:\s(.*)\s\$/ && ($VERSION = $1);
+
+# Constants describing locations in the external array
+use constant EXT_ATTR => 9;
+
+# reference height for characters as a fraction of the 
+# full height/width of the display
+use constant DEFAULT_CHAR_HEIGHT => 1 / 40;
+
+# Static look up tables mapping AST index to Tk equivalent
+# Note that the "black" and "white" entries are really default
+# background + foreground
+# They should be obtained from the canvas itself
+my @COLOURS = qw( black
+		  white
+		  red
+		  green
+		  blue
+		  cyan
+		  magenta
+		  yellow
+		  orange
+		  chartreuse
+		  springgreen
+		  skyblue
+		  purple
+		  pink
+		  darkgrey
+		  grey
+	       );
+
+# Line styles are
+#    solid, long dashes, dash-dot-dash-dot, dotted and dash-dot-dot-dot.
+my @LINE_STYLES = (undef,
+		   undef,  # default is solid
+		   '-',
+		   '-.',
+		   '..',
+		   '-...',
+		  );
+
+# Look up from AST attribute number to Attribute hash key
+my %GRF_PRIM = (
+		&Starlink::AST::Grf::GRF__LINE() => 'LINE',
+		&Starlink::AST::Grf::GRF__MARK() => 'MARK',
+		&Starlink::AST::Grf::GRF__TEXT() => 'TEXT',
+	       );
+
+# Similarly for style
+my %GRF_STYLE = (
+		 &Starlink::AST::Grf::GRF__STYLE() => 'STYLE',
+		 &Starlink::AST::Grf::GRF__WIDTH() => 'WIDTH',
+		 &Starlink::AST::Grf::GRF__SIZE() => 'SIZE',
+		 &Starlink::AST::Grf::GRF__FONT() => 'FONT',
+		 &Starlink::AST::Grf::GRF__COLOUR() => 'COLOUR',
+		);
+
+# Text Anchors
+# points on a compass except no rotation is allowed
+my @AnchorPoints = qw/ n ne e se s sw w ne /;
+my @AnchorAng    = qw/ 0 45 90 135 180 225 270 315 /;
+my %Text_Anchors = (
+		    'CC' => 'center',
+		    'CL' => 'w',
+		    'CR' => 'e',
+
+		    'TC' => 'n',
+		    'TL' => 'nw',
+		    'TR' => 'ne',
+
+		    'BC' => 's',
+		    'BL' => 'sw',
+		    'BR' => 'se',
+		   );
+
 
 =head1 METHODS
 
-=over 
+=over 4
 
 =item B<_GFlush>
 
@@ -64,7 +141,7 @@ any pending graphics to the output device.
 =cut
 
 sub _GFlush {
-   my $external = shift;   
+   my $external = shift;
    my $canvas = $$external[0];
    $canvas->update();
    return 1;
@@ -86,19 +163,39 @@ sub _GLine {
    #print "\n# _GLine()\n";
    #print Dumper( $xf ) . "\n";
    #print Dumper( $yf ) . "\n";
-         
-   if( scalar(@$xf) > 1 && scalar(@$xf) == scalar(@$yf) ) {      
-       
-      my ( @x, @y, @points);
-      foreach my $i ( 0 ... $#$xf ) {
-           
-         ( $x[$i], $y[$i] ) =  _CooTranslate( $external, $$xf[$i], $$yf[$i]);
-         push @points, $x[$i];
-         push @points, $y[$i];
-      }
-      
-      $canvas->createLine( @points );
-   }   
+
+   if( scalar(@$xf) > 1 && scalar(@$xf) == scalar(@$yf) ) {
+
+     # convert GRAPHICS coordinates to pixel coordinates
+     my @points = map {
+       _CooTranslate( $external, $xf->[$_], $yf->[$_]);
+     } ( 0.. $#$xf );
+
+     # modifiers
+     my %opts;
+
+     # Now add additional style information
+     # colour
+     %opts = _attr_to_colour( $canvas, $external->[EXT_ATTR], 'LINE' );
+
+     # line style
+     my $lindex = $external->[EXT_ATTR]->{LINE}->{STYLE};
+     my $lstyle;
+     $lstyle = $LINE_STYLES[ $lindex ] if defined $lindex;
+     $opts{'-dash'} = $lstyle if defined $lstyle;
+
+     # a line width is in units of 1/200 in according to the PGPLOT
+     # standard
+     my $astwid = $external->[EXT_ATTR]->{LINE}->{WIDTH};
+     my $width;
+     $width = ( $astwid / 200 ) . 'i' if defined $astwid;
+     $opts{'-width'} = $width if defined $width;
+
+     # and draw the line
+     my $t = $canvas->createLine( @points, %opts );
+     $canvas->addtag( 'ASTGLine', 'withtag', $t);
+
+   }
    return 1;
 }
 
@@ -121,58 +218,131 @@ sub _GMark {
    #print "\n# _GMark()\n";
    #print Dumper( $xf ) . "\n";
    #print Dumper( $yf ) . "\n";
-         
-   if( scalar(@$xf) > 1 && scalar(@$xf) == scalar(@$yf) ) {      
-        
-      my ( @x, @y, @points);
-      foreach my $i ( 0 ... $#$xf ) {
-         
-         ( $x[$i], $y[$i] ) =  _CooTranslate( $external, $$xf[$i], $$yf[$i]);
-                  
-         # basic scaling factor
-         my $scale = 0.5;
-         
-         # scaling for rectangles
-         if ( $type == 0 || $type == 6 ) {
-           $scale = $scale;
-         } elsif ( $type == 19 ) {
-           $scale = 2*$scale;
-           
-         # scaling fpor circles  
-         } elsif ( $type == 20 ) {
-           $scale = $scale; 
-         } elsif ( $type == 21 ) {
-           $scale = 2*$scale; 
-         } elsif ( $type == 22 ) {
-           $scale = 3*$scale; 
-         } elsif ( $type == 23 ) {
-           $scale = 4*$scale; 
-         } elsif ( $type == 24 ) {
-           $scale = 5*$scale; 
-         } elsif ( $type == 25 ) {
-           $scale = 6*$scale; 
-         } elsif ( $type == 26 ) {
-           $scale = 7*$scale; 
-         } elsif ( $type == 27 ) {
-           $scale = 9*$scale; 
-         }    
-         
-         my $x1 = $x[$i] - $scale*Starlink::AST::Grf::GRF__SIZE();
-         my $y1 = $y[$i] - $scale*Starlink::AST::Grf::GRF__SIZE();
-         my $x2 = $x[$i] + $scale*Starlink::AST::Grf::GRF__SIZE();
-         my $y2 = $y[$i] + $scale*Starlink::AST::Grf::GRF__SIZE(); 
 
-         # RECTANGLE
-         if ( ($type == 0) || ($type == 6) || ($type == 19) ) {
-            $canvas->createRectangle( $x1, $y1, $x2, $y2 );
-         
-         # CIRCLE
-         } if ( $type >= 20 && $type <= 27 ) {
-            $canvas->createOval( $x1, $y1, $x2, $y2 );
-         
-         }
+   if( scalar(@$xf) && scalar(@$xf) == scalar(@$yf) ) {
 
-      }
+     my %opts;
+     %opts = _attr_to_colour( $canvas, $external->[EXT_ATTR], 'MARK');
+
+     # Get the symbol size in pixels
+     my $size = _char_height( $canvas, $external->[EXT_ATTR], 'MARK');
+
+     # and work out how the symbol should be displayed
+
+     # scale factor for some of the types
+     my $scale = 1;
+
+     my $prim;
+     my $fill = 0;
+
+     # Set parameters for duplicated symbols
+     if ($type == 0 || $type == 6 || $type == 19 || $type == 16) {
+       # squares of all types
+       $prim = 'square';
+       if ($type == 19) {
+	 $scale = 1.5;
+       } elsif ($type == 16) {
+	 $scale = 0.5;
+	 $fill = 1;
+       }
+     } elsif ($type == 1 || $type == 4 || $type == 17 ||
+	      ( $type >= 20 && $type <= 27)) {
+       $prim = 'circle';
+
+       if ( $type == 1 ) {
+	 $scale = 0.1;
+       } elsif ($type == 4 || $type == 23) {
+	 $scale = 1;
+       } elsif ($type == 17) {
+	 $scale = 0.75;
+	 $fill = 1;
+       } elsif ($type == 20) {
+	 $scale = 0.25;
+       } elsif ($type == 21) {
+	 $scale = 0.5;
+       } elsif ($type == 22) {
+	 $scale = 0.75;
+       } elsif ($type == 24) {
+	 $scale = 1.5;
+       } elsif ($type == 25) {
+	 $scale = 1.75;
+       } elsif ($type == 26) {
+	 $scale = 2.0;
+       } elsif ($type == 27) {
+	 $scale = 3.0;
+       }
+
+     } elsif ($type == 2 ) {
+       # cross
+       $prim = 'cross';
+     } else {
+       # do not yet know how to render this. Default
+       $prim = 'circle';
+       $scale = 1;
+       $fill = 1;
+     }
+
+     # Configure the fill color
+     if (!$fill) {
+       # by default the color to opt code uses -fill
+       # but we need to use -outline if fill is not specified
+       if (exists $opts{'-fill'}) {
+	 $opts{'-outline'} = $opts{'-fill'};
+	 delete $opts{'-fill'};
+       }
+     }
+
+     foreach my $i ( 0 ... $#$xf ) {
+
+       # convert to canvas coordinates
+       my ( $x, $y ) =  _CooTranslate( $external, $xf->[$i], $yf->[$i]);
+
+       # now draw them
+       my $item;
+       if ($prim eq 'square') {
+
+	 my $dist = $size * $scale / 2;
+	 my $x1 = $x - $dist;
+	 my $x2 = $x + $dist;
+	 my $y1 = $y - $dist;
+	 my $y2 = $y + $dist;
+
+	 $item = $canvas->createRectangle( $x1, $y1, $x2, $y2, %opts );
+
+       } elsif ($prim eq 'circle') {
+
+	 my $dist = $size * $scale / 2;
+	 my $x1 = $x - $dist;
+	 my $x2 = $x + $dist;
+	 my $y1 = $y - $dist;
+	 my $y2 = $y + $dist;
+
+	 $item = $canvas->createOval( $x1, $y1, $x2, $y2, %opts );
+
+
+       } elsif ($prim eq 'cross') {
+
+	 # we may have to respect line widths
+	 my $dist = $size * $scale / 2;
+	 my $x1 = $x - $dist;
+	 my $x2 = $x + $dist;
+	 my $y1 = $y - $dist;
+	 my $y2 = $y + $dist;
+
+	 $item = $canvas->createLine( $x, $y1, $x, $y2, %opts );
+	 $canvas->addtag( 'ASTGMark', 'withtag', $item);
+	 $item = $canvas->createLine( $x1, $y, $x2, $y, %opts );
+
+
+
+       } else {
+	 # should not happen
+	 ReportGrfError("_GMark: Bizarre inability to determine symbol type");
+	 return (0);
+       }
+
+       $canvas->addtag( 'ASTGMark', 'withtag', $item);
+     }
    }
    return 1;
 }
@@ -205,10 +375,18 @@ world coordinates. If necessary the supplied value should be negated to
 ensure that positive values always refer to displacements from  bottom to
 top on the screen.
 
+Note that we match the PGPLOT definition of default character height.
+ie the default character height one-fortieth of the height or width of
+the view surface (whichever is less).
+
+An optional argument can be provided at the end of the argument list
+(AST will not do this). This should be a ref to a scalar. On return,
+it will contain the object that was just plotted.
+
 =cut
 
 sub _GText {
-   my ( $external, $text, $xf, $yf, $just, $upx, $upy ) = @_;
+   my ( $external, $text, $xf, $yf, $just, $upx, $upy, $ref ) = @_;
    my $canvas = $$external[0];
    #print "# _GText: Placeholder routine called\n";
 
@@ -216,19 +394,63 @@ sub _GText {
    #print "\n# _GText( $text )\n";
    #print Dumper( $xf ) . "\n";
    #print Dumper( $yf ) . "\n";
-      
+
+   print "$text : $just \n" if $text =~ /decl/i;
+
    if( defined $text && length($text) != 0 ) {
-      
-      # multiple the current co-ordinate
-      my ( $x, $y ) =  _CooTranslate($external, $xf, $yf);   
-      
-      # draw text
-      $canvas->createText( $x, $y, -text => $text );
+
+     # work out the X and Y anchoring
+     my $xa = substr($just,1,1) || 'C';
+     my $ya = substr($just,0,1) || 'C';
+
+     # and reconstruct it
+     $just = $ya. $xa;
+
+     # Try to fudge the most common rotation
+     if ($upx == 1 && $upy == 0 ) {
+       # massage to attempt vertical text
+       $text = join("\n", split(//, $text));
+     } elsif ($upx == -1 && $upy == 0 ) {
+       # massage to attempt vertical text
+       $text = join("\n", reverse split(//, $text));
+     }
+
+     # option handling
+     my %opts;
+     %opts = _attr_to_colour( $canvas, $external->[EXT_ATTR], 'TEXT');
+
+     # translate to the pixel coordinates
+     my ( $x, $y ) =  _CooTranslate($external, $xf, $yf);
+
+     # Specify text anchor
+     $opts{'-anchor'} = ( exists $Text_Anchors{$just} ?
+			  $Text_Anchors{$just} : $Text_Anchors{CC});
+
+
+     # apply scaling
+     my $charh = _char_height( $canvas, $external->[EXT_ATTR], 'TEXT');
+
+     # select the font object
+     my $fi = $external->[EXT_ATTR]->{TEXT}->{FONT};
+     my $font = $external->[EXT_ATTR]->{FONTS}->[$fi];
+
+     # size in pixels is specified as negative number
+     $font->configure( '-size' => (-1 * $charh ) );
+
+     # draw text
+     my $item = $canvas->createText( $x, $y, -text => $text, %opts,
+				     -font => $font,
+				   );
+     $canvas->addtag( 'ASTGText', 'withtag', $item);
+
+     # if we have a scalar ref
+     $$ref = $item if defined $ref;
+
    }
-   
+
    # Return, all is well strangely
-   return 1;   
-}            
+   return 1;
+}
 
 
 =item B<_GScales>
@@ -250,19 +472,19 @@ sub _GScales {
     my ($xglo,$xghi,$yglo,$yghi) = @$external[1 .. 4];
     my ($xplo,$xphi,$yplo,$yphi) = @$external[5 .. 8];
     my ($xmin,$xmax,$ymin,$ymax) = _CooBox($external);
-        
+
     my ($nx1, $nx2, $ny1, $ny2);
     my ($wx1, $wx2, $wy1, $wy2);
-    
+
     $nx1 = $xmin;
     $nx2 = $xmax;
     $ny1 = $ymax;
     $ny2 = $ymin;
-    
+
     $wx1 = $xglo;
     $wx2 = $xghi;
     $wy1 = $yghi;
-    $wy2 = $yglo;    
+    $wy2 = $yglo;
 
     if( $wx2 != $wx1 && $wy2 != $wy1 && $nx2 != $nx1 && $ny2 != $ny1 ) {
        $alpha = ( $nx2 - $nx1 ) / ( $wx2 - $wx1 );
@@ -271,8 +493,8 @@ sub _GScales {
        ReportGrfError("_GScales: The graphics window has zero size");
        return (0);
     }
-    return ( 1, $alpha, $beta );   
-}       
+    return ( 1, $alpha, $beta );
+}
 
 
 =item B<_GTxExt>
@@ -281,7 +503,7 @@ This function returns the corners of a box which would enclose the
 supplied character string if it were displayed using astGText. The 
 returned box INCLUDES any leading or trailing spaces.
 
-   my ( $status, $xb, $yb ) = _GTxtExt( $w, $text, $x, $y, $just, $upx, $upy);
+   my ( $status, $xb, $yb ) = _GTxExt( $w, $text, $x, $y, $just, $upx, $upy);
 
 where $x is the reference x coordinate, $y is the reference y coordinate, 
 and where $justification is a character string which specifies the
@@ -321,25 +543,44 @@ Notes:
 =cut
 
 sub _GTxExt {
-   my ( $external, $text, $x, $y, $just, $upx, $upy ) = @_;
-   my $canvas = $$external[0];
-   #print "# _GTxExt: Placeholder routine called\n";
-   
-   # initalise @$xb and @$yb
-   my ( @xb, @yb );
-   $xb[0] = $x;
-   $yb[0] = $y;
-   $xb[1] = $x + (length($text)*12)/( $canvas->cget( '-width' ) );
-   $yb[1] = $y;
-   $xb[2] = $x + (length($text)*12)/( $canvas->cget( '-width' ) );
-   $yb[2] = $y + 12/( $canvas->cget( '-height' ) );
-   $xb[3] = $x;
-   $yb[3] = $y + 12/( $canvas->cget( '-height' ) );
-    
-   # Return
-   return (1, \@xb, \@yb );     
-   
-}          
+  my $external = shift;
+  my $canvas = $external->[0];
+
+  # The easy plan is to:
+  #   get GText to plot the text
+  #   Attach a new tag to it
+  #   read the bounding box
+  #   delete the item from the canvas
+  my $item;
+  _GText( $external, @_, \$item );
+
+  # add a unique tag
+  $canvas->addtag('ASTGTxExt_TEMP', 'withtag', $item);
+
+  # read the bounding box
+  my @bbox = $canvas->bbox( 'ASTGTxExt_TEMP' );
+
+  # delete the text
+  $canvas->delete( 'ASTGTxExt_TEMP' );
+
+  # to convert these coordinates back to the correct units
+  # we need the width and height of the canvas
+  my $width = $canvas->cget( '-width' );
+  my $height = $canvas->cget( '-height' );
+
+  # convert to AST form
+  my @xb = @bbox[0,2,2,0];
+  my @yb = @bbox[1,1,3,3];
+
+  @xb = map { $_ / $width } @xb;
+  @yb = map { 1 - $_ / $height } @yb;
+
+  #use Data::Dumper;
+  #print Dumper([$_[0], \@xb, \@yb]);
+
+  # Return
+  return (1, \@xb, \@yb);
+}
 
 =item B<_GQch>
 
@@ -348,11 +589,11 @@ horizontally in world coordinates.
 
    my ( $status, $chv, $chh ) = _GQch( $w );
 
-Where $chv is a reference which is to receive the height of characters 
-drawn with a vertical baseline. This will be an increment in the X axis.
+Where $chv is the height of characters drawn with a vertical
+baseline. This will be an increment in the X axis.  Where $chh is the
+height of characters drawn with a horizontal baseline. This will be an
+increment in the Y axis.
 
-Where $chh is a reference which is to receive the height of characters 
-drawn with a horizontal baseline. This will be an increment in the Y axis.
 
 =cut
 
@@ -360,11 +601,14 @@ sub _GQch {
    my $external = shift;
    my $canvas = $$external[0];
    #print "# _GQch: Placeholder routine called\n";
-   
-   my $chv = 12/$canvas->cget( '-height' );
-   my $chh = 12/$canvas->cget( '-width' );  
-   return (1, $chv, $chh );  
-}   
+
+   # the Tk definition of world coordinates (from AST viewpoint)
+   # is fraction of viewport
+   my $charh = _char_height( $canvas, $external->[EXT_ATTR], 'TEXT' );
+   my $chv = $charh / $canvas->cget( '-width' );
+   my $chh = $charh / $canvas->cget( '-height' );
+   return (1, $chv, $chh );
+}
 
 
 =item B<_GAttr>
@@ -399,47 +643,39 @@ AST's grf.h:
 
 =cut
 
-my @gattrs; # Global
 sub _GAttr {
    my ( $external, $att, $val, $prim ) = @_;
-   #print "# _GAttr: Placeholder routine called\n";
 
-  my $MAX_ATTR = 5;
-  my $i;
-  if ($att == &Starlink::AST::Grf::GRF__STYLE() ) {
-    $i = 1;
-  } elsif ( $att == &Starlink::AST::Grf::GRF__WIDTH() ) {
-    $i = 2;
-  } elsif ( $att == &Starlink::AST::Grf::GRF__SIZE() ) {
-    $i = 3;
-  } elsif ( $att == &Starlink::AST::Grf::GRF__FONT() ) {
-    $i = 4;
-  } elsif ( $att == &Starlink::AST::Grf::GRF__COLOUR() ) {
-    $i = 5;
-  } else {
-    print "# Bad ATT value: ", $att ."\n";
-  }
+   # Get the attribute hash
+   my $attr = $external->[EXT_ATTR];
 
-  my $j;
-  if ($prim == &Starlink::AST::Grf::GRF__LINE() ) {
-    $j = 1;
-  } elsif ($prim == &Starlink::AST::Grf::GRF__MARK() ) {
-    $j = 2;
-  } elsif ($prim == &Starlink::AST::Grf::GRF__TEXT() ) {
-    $j = 3;
-  } else {
-    print "# Bad PRIM value: $prim\n";
-  }
+   # possible return value
+   my $old_value = undef;
 
-  # Store the new value if required
-  # Convert prim and att to index of 2d array
-  my $index = ( $MAX_ATTR * ($att - 1) + $prim );
-  my $old = $gattrs[$index];
-  $old = &Starlink::AST::AST__BAD() if !defined $old;
-  $gattrs[$index] = $val if $val != &Starlink::AST::AST__BAD();
+   # determine which primitive we are dealing with
+   my $PRIM = $GRF_PRIM{$prim};
+   if (!defined $PRIM) {
+     ReportGrfError("_GAttr: Unknown primitive: $prim\n");
+     return ( 0 );
+   }
 
-  return (1, $old);
-}   
+   # and which style
+   my $STYLE = $GRF_STYLE{$att};
+   if (!defined $STYLE) {
+     ReportGrfError("_GAttr: Unknown style: $att\n");
+     return ( 0 );
+   }
+
+   # Now process the style request, relying on each plotting system
+   # to be able to work out what these numbers mean
+   $old_value = $attr->{$PRIM}->{$STYLE};
+   $attr->{$PRIM}->{$STYLE} = $val if $val != &Starlink::AST::AST__BAD();
+
+   # map to bad value if appropriate
+   $old_value = &Starlink::AST::AST__BAD() if !defined $old_value;
+
+   return (1, $old_value);
+}
 
 
 =item B<_GCap>
@@ -497,7 +733,7 @@ sub _GCap {
 # Internal error setting routine
 sub ReportGrfError {
   my $text = shift;
-  warn "Generated AST error in perl PGPLOT callback: $text\n";
+  warn "Generated AST error in perl Tk callback: $text\n";
   Starlink::AST::_Error( &Starlink::AST::Status::AST__GRFER(), $text);
 }
 
@@ -525,6 +761,9 @@ sub _CooBox {
    return ($xleft,$xright,$ybottom,$ytop);
 }
 
+# convert GRAPHICS coordinates (0,1)
+# to canvas pixel coordinates
+
 sub _CooTranslate {
    my ($external, $xf, $yf) = @_;
    my $canvas = $$external[0];
@@ -546,11 +785,121 @@ sub _CooTranslate {
    return ( $x, $y );
 }   
 
+
+# Given the attributes hash and the primitive type,
+# return a hash with the correct arguments for the canvas
+# color settings
+
+sub _attr_to_colour {
+  my ($canvas, $external, $PRIM) = @_;
+
+  my %opts;
+  my $ci = $external->{$PRIM}->{COLOUR};
+  if (defined $ci && $ci >= 0 && $ci <= $#COLOURS) {
+    # background color
+    if ($ci == 0) {
+      $opts{'-fill'} = $canvas->cget('-background');
+    } elsif ($ci == 1) {
+      # foreground color
+      my $bg = $canvas->cget('-background');
+      my $fg = $COLOURS[$ci];
+
+      if ($bg eq $fg) {
+	# need to choose something else
+	$fg = $COLOURS[0];
+      };
+
+      $opts{'-fill'} = $fg;
+    } else {
+      $opts{'-fill'} = $COLOURS[$ci]
+    }
+  }
+  return %opts;
+}
+
+# Calculate the default character height
+
+# To match PGPLOT we define the default character height one-fortieth
+# of the height or width of the view surface (whichever is less).
+
+# Size specifications are then relative to this unit
+
+# this height is returned in pixels
+
+sub _def_char_height {
+  my $canvas = shift;
+  my $width = $canvas->cget('-width');
+  my $height = $canvas->cget('-height');
+
+  my $wh = $width * DEFAULT_CHAR_HEIGHT;
+  my $hh = $height * DEFAULT_CHAR_HEIGHT;
+
+  return ( $wh < $hh ? $wh : $hh );
+}
+
+# Return the character height in pixels for this particular primitive
+# type.
+# Requires the canvas, attr hash and the primitive name (TEXT,MARK,LINE)
+
+sub _char_height {
+  my ($canvas, $attr, $prim) = @_;
+
+  my $dch = _def_char_height( $canvas );
+  my $scale = $attr->{$prim}->{SIZE};
+  $scale = 1 unless defined $scale;
+  return ($dch * $scale);
+}
+
+# Routine to initialise the attributes. This includes the creation
+# of Font objects that match the desired PGPLOT types
+
+sub _init_canvas_attrs {
+  my ($canvas, $attr) = @_;
+
+  # prefill size attr to 1 since that is what PGPLOT does
+  $attr->{TEXT}->{SIZE} = 1;
+  $attr->{TEXT}->{FONT} = 1;
+  $attr->{MARK}->{SIZE} = 1;
+
+  # This will be a lookup just like the Colour array except that
+  # it has to be per-widget rather than global
+  my @fonts;
+
+  # Simple font
+  $fonts[1] = $canvas->Font( -family => 'Courier',
+			     -slant => 'roman',
+			     -size => 12,
+			   );
+
+  # roman font
+  $fonts[2] = $canvas->Font( -family => 'Times',
+			     -slant => 'roman',
+			     -size => 12,
+			   );
+
+  # italic
+  $fonts[3] = $canvas->Font( -family => 'Time',
+			     -slant => 'italic',
+			     -size => 12,
+			   );
+
+  # script
+  $fonts[4] = $canvas->Font( -family => 'Helvetica',
+			     -slant => 'italic',
+			     -size => 12,
+			   );
+
+  $attr->{FONTS} = \@fonts;
+
+}
+
 =back
 
 =head1 COPYRIGHT
 
-Copyright (C) 2004 University of Exeter. All Rights Reserved.
+Copyright (C) 2004 University of Exeter.
+Copyright (C) 2005 Particle Physics and Astronomy Research Council.
+All Rights Reserved.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -568,6 +917,7 @@ Place,Suite 330, Boston, MA  02111-1307, USA
 =head1 AUTHORS
 
 Alasdair Allan E<lt>aa@astro.ex.ac.ukE<gt>,
+Tim Jenness E<lt>t.jenness@jach.hawaii.eduE<gt>
 
 =cut
 
@@ -581,12 +931,20 @@ use Starlink::AST::Tk;
 sub tk {
   my $self = shift;
   my $canvas = shift;
- 
+
+  # Store the information we need to pass to the plot functions
   my @external;
   push @external, $canvas;
   push @external, $self->GBox();
   push @external, $self->PBox();
-  
+
+  my $attr = {};
+  push @external, $attr; # hash for attributes
+
+  # initialise the attribute hash
+  Starlink::AST::Tk::_init_canvas_attrs( $canvas, $attr );
+
+  # register callbacks
   $self->GExternal( \@external );
   $self->GFlush(\&Starlink::AST::Tk::_GFlush);  
   $self->GLine(\&Starlink::AST::Tk::_GLine);
